// This demo and test program shows how to use the Curve Generator tool
// with nonlinear sensors.
// It is based on the E-LAB EvaBoard and the E-LAB NonLinSensor Board
// The schematics of the sensor board "NonLinSensSch.PDF" is included in the
// DOCs section of the installation.
//
// Please note that the analog circuit of this board is not intended for using
// in a real application. The voltage span of the two temperature sensor is too
// low for precise measurement. The distance meter and the lumen meter can be
// used without any changes because their output voltage swing is large enough.
//
// Furthermore the two temperature sensors are very non-linear which can not be
// directly seen in the CurveGen tool's graph. Extending the temperature span
// from -50C to +200C then will show this.
//
// The lumen meter is very sensitive to 50/60Hz lamp distortion. In order to
// remove this "noise" a low-pass filter should be used. A notch filter can
// also supress the overlayed sine, but because the used sensor is very slow
// a low pass filter is sufficient.
//
// The distance sensor is the Sharp type GP2D120
// The light sensor is a LDR07 type
// The first temperature sensor is a silicon temperature sensor KTY81
// The second temperature sensor is a platinum type with 1kOhm called PT1000
//
// In oder to save CPU port pins and ADC channels the sonsor board has an onboard
// analog multiplexer HC4051 so only one ADC channel is used for the sensors
// The multiplexer is controlled in the callback function "OnADCread".

program NonLinear;

{$NOSHADOW}

Device = mega128, VCC = 5;

Define_Fuses
  LockBits0      = [];
  FuseBits0      = [];
  FuseBits1      = [CKOPT, SPIEN, JTAGEN, OCDEN];
  FuseBits2      = [];
  ProgMode       = JTAG;

Import SysTick, SerPort, ADCPort, LCDmultiPort, BeepPort, TWImaster;

From System Import LongWord, LongInt, Float;

Define
  ProcClock      = 16000000;        {Hertz}
  SysTick        = 5;               {msec}
  StackSize      = $0064, iData;
  FrameSize      = $0064, iData;
  SerPort        = 19200, Stop2;    {Baud, StopBits|Parity}
  RxBuffer       = 32, iData;
  TxBuffer       = 32, iData;
  ADCchans       = 4, iData;
  ADCpresc       = 128;
  TWIpresc       = TWI_BR400;
  LCDmultiPort   = I2C_TWI;
  LCDrows_M      = 4;               {rows}
  LCDcolumns_M   = 16;              {columns per line}
  LCDtype_M      = 44780;
  BeepPort       = PortG, 4;

Implementation

{$IDATA}

{--------------------------------------------------------------}
{ Type Declarations }

type


{--------------------------------------------------------------}
{ Const Declarations }
const
  SharpLookUp   : array[1..(100 * 4) + 3] of byte = 'Sharp.crvg';
  KTYLookUp     : array[1..(100 * 4) + 3] of byte = 'KTY.crvg';
  PT1000LookUp  : array[1..(100 * 4) + 3] of byte = 'PT1000.crvg';
  LDRLookUp     : array[1..(100 * 4) + 3] of byte = 'LDR.crvg';
  // all tables consist of 100 point.x point.y of integer, 3bytes info


{--------------------------------------------------------------}
{ Var Declarations }
{$IDATA}
var
  ii, xx, yy       : integer;
  LDRint           : integer;
  ADC_Arr          : array[0..7] of integer;
  

{--------------------------------------------------------------}
{ functions }
// This is a CallBack function out of the SysTick!
// So there are some restrictions
// 1. Make it very short and fast otherwise the system can get into troubles
// 2. Only use ACCB/R16, ACCA/R17, ACCALO/R18, ACCCLO/R30 and ACCCHI/R31
procedure onADCread;
begin
  // ADC result as word value in ACCB/R16 = loByte, ACCA/R17 = hiByte
  // current ADC channel in ACCALO/R18
  // In this app ADC channel0 has an external multiplexer with 8 channel selector
  // the selector is driven by PortD.4 .. PortD.6
  ASM;
    ANDI      _ACCALO, 7      ; if ADC chan = 0 then we must place the result into the selector array
    BRNE      ADCrd_X         ; nothing to do
                              ; we must build a pointer into the array
    IN        _ACCALO, PortD  ; get the current selector position
    ANDI      _ACCALO, 70h    ; separate channel select bits
    LSR       _ACCALO         ; justify it and get the word access offset
    LSR       _ACCALO         ;
    LSR       _ACCALO         ;
    LDI       _ACCCLO, NonLinear.ADC_Arr AND 0FFh
    LDI       _ACCCHI, NonLinear.ADC_Arr SHRB 8
    ADD       _ACCCLO, _ACCALO
    BRCC      ADCrd_I
    INC       _ACCCHI
    ADCrd_I:
    ST        Z, _ACCB        ; store ADC result into the array
    STD       Z+1, _ACCA
                              ; now we have to set the selectors next position
    IN        _ACCALO, PortD  ; get the current selector position
    ADDI      _ACCALO, 10h    ; next external channel
    ANDI      _ACCALO, 70h    ; separate channel select bits
    IN        _ACCA, PortD    ; get the current selector position
    CBR       _ACCA, 70h      ; remove previous channel select bits
    OR        _ACCA, _ACCALO  ; insert chan select bits
    OUT       PORTD, _ACCA    ; update external MUX
    
    ADCrd_X:
  ENDASM;
end;

procedure PortInit;
begin
  DDRD:= DDRD or $70;         // external ADC MUX select bits
  PortD:= PortD and $8F;      // start with ext chan 0
end;

{--------------------------------------------------------------}
{ Main Program }
{$IDATA}

begin
  PortInit;
  EnableInts;

  BeepSiren(0, 3);
  mDelay(2000);

  LCDsetup_M(LCD_m1);
  LCDcursor_M(LCD_m1, true, false);
  Write(LCDOut_M,  'E-LAB NonLinear');
  LCDxy_M(LCD_m1, 0, 1);
  Write(LCDOut_M,  'Demo program');
  BeepOut(2500, 20);
  mDelay(2000);

  loop
    xx:= ADC_Arr[0];                                 // in real world a 50Hz (60Hz) filter is needed
    LDRint:= ((LDRint * 20) + xx) div 21;            // integrate 50Hz light overlay

(* only show the raw ADC results here
    LCDxy_M(LCD_m1, 0, 1);
    Write(LCDout_M, 'LDR    : '+IntToStr(LDRint));
    LCDClrEol_M(LCD_M1);

    xx:= ADC_Arr[1];
    LCDxy_M(LCD_m1, 0, 2);
    Write(LCDout_M, 'KTY    : '+IntToStr(xx));
    LCDClrEol_M(LCD_M1);

    xx:= ADC_Arr[2];
    LCDxy_M(LCD_m1, 0, 3);
    Write(LCDout_M, 'PT1000 : '+IntToStr(xx));
    LCDClrEol_M(LCD_M1);
*)
(* only show the light sensor relevant data
    yy:= LDRint;                                     // get LDR light sensor
    xx:= round(Float(yy) * 4.887);
    LCDxy_M(LCD_m1, 0, 1);
    Write(LCDout_M, 'Inp volt : '+IntToStr(xx:5:3));
    LCDClrEol_M(LCD_M1);
    LCDxy_M(LCD_m1, 0, 2);
    Write(LCDout_M, 'Inp val  : '+IntToStr(yy:4));
    LCDClrEol_M(LCD_M1);
    if InterPolY(@LDRLookUp, yy, ii) then
      LCDxy_M(LCD_M1, 0, 3);
      Write(LCDout_M, 'Lumen Lx : '+IntToStr(ii:4));
      LCDClrEol_M(LCD_M1);
    else
      LCDxy_M(LCD_M1, 0, 3);
      Write(LCDout_M, 'Out of scope');
      LCDClrEol_M(LCD_m1);
    endif;
*)
(* only show the KTY temp sensor relevant data
    yy:= ADC_Arr[1];                        // get KTY temperature sensor
    xx:= round(Float(yy) * 4.887);
    LCDxy_M(LCD_m1, 0, 1);
    Write(LCDout_M, 'Inp volt : '+IntToStr(xx:5:3));
    LCDClrEol_M(LCD_M1);
    LCDxy_M(LCD_m1, 0, 2);
    Write(LCDout_M, 'Inp val  : '+IntToStr(yy:4));
    LCDClrEol_M(LCD_M1);
    if InterPolY(@KTYLookUp, yy, ii) then
      LCDxy_M(LCD_M1, 0, 3);
      Write(LCDout_M, 'Temp deg : '+IntToStr(ii:4:1));
      LCDClrEol_M(LCD_M1);
    else
      LCDxy_M(LCD_M1, 0, 3);
      Write(LCDout_M, 'Out of scope');
      LCDClrEol_M(LCD_m1);
    endif;
*)
(* only show the PT1000 temp sensor relevant data
    yy:= ADC_Arr[2];                        // get PT1000 temperature sensor
    xx:= round(Float(yy) * 4.887);
    LCDxy_M(LCD_m1, 0, 1);
    Write(LCDout_M, 'Inp volt : '+IntToStr(xx:5:3));
    LCDClrEol_M(LCD_M1);
    LCDxy_M(LCD_m1, 0, 2);
    Write(LCDout_M, 'Inp val  : '+IntToStr(yy:4));
    LCDClrEol_M(LCD_M1);
    if InterPolY(@PT1000LookUp, yy, ii) then
      LCDxy_M(LCD_M1, 0, 3);
      Write(LCDout_M, 'Temp deg : '+IntToStr(ii:4:1));
      LCDClrEol_M(LCD_M1);
    else
      LCDxy_M(LCD_M1, 0, 3);
      Write(LCDout_M, 'Out of scope');
      LCDClrEol_M(LCD_m1);
    endif;
*)
(* only show the distance sensor relevant data
    yy:= ADC_Arr[3];                        // get Sharp distance sensor
    xx:= round(Float(yy) * 4.887);
    LCDxy_M(LCD_m1, 0, 1);
    Write(LCDout_M, 'Inp volt : '+IntToStr(xx:5:3));
    LCDClrEol_M(LCD_M1);
    LCDxy_M(LCD_m1, 0, 2);
    Write(LCDout_M, 'Inp val  : '+IntToStr(yy:4));
    LCDClrEol_M(LCD_M1);
    if InterPolY(@SharpLookUp, yy, ii) then
      LCDxy_M(LCD_M1, 0, 3);
      Write(LCDout_M, 'Dist  cm : '+IntToStr(ii:4:1));
      LCDClrEol_M(LCD_M1);
    else
      LCDxy_M(LCD_M1, 0, 3);
      Write(LCDout_M, 'Out of scope');
      LCDClrEol_M(LCD_m1);
    endif;
*)

    yy:= LDRint;                            // get LDR light sensor
    LCDxy_M(LCD_M1, 0, 0);
    if InterPolY(@LDRLookUp, yy, ii) then
      Write(LCDout_M, 'Lumen Lx : '+IntToStr(ii:4));
    else
      Write(LCDout_M, 'Out of scope');
    endif;
    LCDClrEol_M(LCD_M1);

    yy:= ADC_Arr[1];                        // get KTY temperature sensor
    LCDxy_M(LCD_M1, 0, 1);
    if InterPolY(@KTYLookUp, yy, ii) then
      Write(LCDout_M, 'Temp KTY : '+IntToStr(ii:4:1));
    else
      Write(LCDout_M, 'Out of scope');
    endif;
    LCDClrEol_M(LCD_M1);

    yy:= ADC_Arr[2];                        // get PT1000 temperature sensor
    LCDxy_M(LCD_M1, 0, 2);
    if InterPolY(@PT1000LookUp, yy, ii) then
      Write(LCDout_M, 'Temp PT  : '+IntToStr(ii:4:1));
    else
      Write(LCDout_M, 'Out of scope');
    endif;
    LCDClrEol_M(LCD_M1);

    yy:= ADC_Arr[3];                        // get Sharp distance sensor
    LCDxy_M(LCD_M1, 0, 3);
    if InterPolY(@SharpLookUp, yy, ii) then
      Write(LCDout_M, 'Dist cm  : '+IntToStr(ii:4:1));
    else
      Write(LCDout_M, 'Out of scope');
    endif;
    LCDClrEol_M(LCD_M1);

    mDelay(100);
  endloop;
end NonLinear.

